from eis1600.markdown.re_patterns import MIU_HEADER


class YAMLHandler:
    """A class to take care of the MIU YAML Headers

    :param bool reviewed: Indicating if this record has been manually reviewed. If True, the record will not be
    changed by automated analysis, defaults to False.
    """

    def __init__(self, reviewed: bool = False) -> None:
        self.reviewed = reviewed

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)

    def set_headings(self, headings: dict):
        for key, val in headings.items():
            self.__setattr__(key, val)

    def get_yamlfied(self) -> str:
        yaml_str = MIU_HEADER + '\n\n'
        for key, val in vars(self).items():
            yaml_str += key + '    : ' + str(val) + '\n'
        yaml_str += '\n' + MIU_HEADER + '\n\n'

        return yaml_str



from eis1600.miu.YAMLHandler import YAMLHandler

from eis1600.markdown.re_patterns import MIU_HEADER_PATTERN, NEWLINES_PATTERN


def create_yml_header(headings: dict) -> str:
    """Creates a YAML header for the current MIU file.

    :param dict headings: Dict of headings, which are the super elements of the current MIU.
    :return str: YAML header for the current MIU.
    """

    yml_header = YAMLHandler()
    if headings:
        yml_header.set_headings(headings)

    return yml_header.get_yamlfied()


def extract_yml_header_and_text(miu_file: str, miu_id: str, is_header: bool) -> (str, str):
    """ Returns the YAML header and the text as a tuple.

    Splits the MIU file into a tuple of YAML header and text.
    :param str miu_file: Path to the MIU file from which to extract YAML header and text.
    :param str miu_id: MIU UID needed as a super element of the extracted YAML header.
    :param bool is_header: Indicates if the current MIU is the YAML header of the whole work and if so skips removing
    blank lines.
    :return (str, str): Tuple of the extracted YAML header and text.
    """

    with open(miu_file, 'r', encoding='utf-8') as file:
        text = ''
        miu_yml_header = ''
        print(MIU_HEADER_PATTERN)
        for line in iter(file):
            if MIU_HEADER_PATTERN.match(line):
                # Omit the #MIU#Header# line as it is only needed inside the MIU.EIS1600 file, but not in YMLDATA.yml
                next(file)
                line = next(file)
                miu_yml_header = '#' + miu_id + '\n---\n'
                while not MIU_HEADER_PATTERN.match(line):
                    miu_yml_header += line
                    line = next(file)
                # Skip empty line after #MIU#Header#
                next(file)
            else:
                text += line
            # Replace new lines which separate YAML header from text
            if not is_header:
                text = NEWLINES_PATTERN.sub('\n\n', text)

        return miu_yml_header, text


def update_yml_header():
    # TODO: update YAML header with MIU related information from automated analyses and manual tags
    pass



class HeadingTracker:
    """A class to keep track of the super elements of a MIU.

    This class keeps track of the headings on different levels to keep this information in the MIU YAML header.
    :ivar str level1: Level 1 heading.
    :ivar str level2: Level 2 heading.
    :ivar str level3: Level 3 heading.
    :ivar str level4: Level 4 heading.
    """

    def __init__(self) -> None:
        """Constructor which sets attributes to empty strings."""

        self.level1 = ''
        self.level2 = ''
        self.level3 = ''
        self.level4 = ''

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)

    def get_curr_state(self) -> dict:
        """Get current state of the tacker as dict.

        Returns a dictionary of the current state, but only with attributes whose value is not an empty string.
        :return dict: Dict of the current state of the tracker.
        """

        if not self.level1:
            return None

        headings_dict = {}
        for key, val in self.__dict__.items():
            if val:
                headings_dict[key] = val

        return headings_dict

    def get_yamlfied(self) -> str:
        """Stringifies HeadingTracker in YAML format, only includes levels which are set.

        :return str: returns the HeadingTracker in YAML format as a string.
        """

        if not self.level1:
            return ''

        heading_tracker_str = 'Heading_1    : ' + self.level1 + '\n'
        if self.level2:
            heading_tracker_str += 'Heading_2    : ' + self.level2 + '\n'
            if self.level3:
                heading_tracker_str += 'Heading_3    : ' + self.level3 + '\n'
                if self.level4:
                    heading_tracker_str += 'Heading_4    : ' + self.level4 + '\n'

        return heading_tracker_str

    def track(self, level: int, heading: str) -> None:
        """Checks which of the levels changed and sets all sub levels to an empty string.

        :param int level: The level of the heading indicated by the number of leading `|`.
        :param str heading: The new heading text for the given level.
        """

        if level == 1:
            self.level1 = heading
            self.level2 = ''
            self.level3 = ''
            self.level4 = ''
        elif level == 2:
            self.level2 = heading
            self.level3 = ''
            self.level4 = ''
        elif level == 3:
            self.level3 = heading
            self.level4 = ''
        else:
            self.level4 = heading



from os.path import splitext, split
from typing import Optional

from pathlib import Path

from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.miu.yml_handling import create_yml_header, extract_yml_header_and_text
from eis1600.markdown.re_patterns import HEADER_END_PATTERN, HEADING_PATTERN, MIU_UID_PATTERN, UID_PATTERN


def disassemble_text(infile: str, verbose: Optional[bool] = None) -> None:
    """Disassemble text into MIU files.

    Retrieve MIU files by disassembling the text based on the EIS1600 mARkdown.
    :param str infile: Path to the file which is to be disassembled.
    :param bool verbose: If True outputs a notification of the file which is currently processed, optional.
    """

    heading_tracker = HeadingTracker()
    path, uri = split(infile)
    uri, ext = splitext(uri)
    ids_file = path + '/' + uri + '.IDs'
    miu_dir = Path(path + '/' + uri + '/')
    uid = ''
    miu_text = ''

    if verbose:
        print(f'Disassemble {uri}')

    miu_dir.mkdir(exist_ok=True)
    miu_uri = miu_dir.__str__() + '/' + uri + '.'

    with open(infile, 'r', encoding='utf8') as text:
        with open(ids_file, 'w', encoding='utf8') as ids_tree:
            for text_line in iter(text):
                if HEADER_END_PATTERN.match(text_line):
                    uid = 'header'
                    ids_tree.write(uid + '\n')
                    miu_text += text_line
                    with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                        miu_file.write(miu_text + '\n')
                    miu_text = ''
                    uid = 'preface'
                    next(text)  # Skip empty line after header
                elif MIU_UID_PATTERN.match(text_line):
                    if HEADING_PATTERN.match(text_line):
                        m = HEADING_PATTERN.match(text_line)
                        heading_tracker.track(len(m.group('level')), m.group('heading'))
                    with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                        miu_file.write(miu_text)
                    uid = UID_PATTERN.match(text_line).group('UID')
                    ids_tree.write(uid + '\n')
                    miu_text = create_yml_header(heading_tracker.get_curr_state())
                    miu_text += text_line
                else:
                    miu_text += text_line
            # last MIU needs to be written to file when the for-loop is finished
            with open(miu_uri + uid + '.EIS1600', 'w', encoding='utf8') as miu_file:
                miu_file.write(miu_text)


def reassemble_text(infile, verbose):
    path, uri = split(infile)
    uri, ext = splitext(uri)
    file_path = path + '/' + uri
    ids = []

    if verbose:
        print(f'Reassemble {uri}')

    with open(file_path + '.IDs', 'r', encoding='utf-8') as ids_file:
        ids.extend([line[:-1] for line in ids_file.readlines()])

    with open(file_path + '.EIS1600', 'w', encoding='utf-8') as text_file:
        with open(file_path + '.YMLDATA.yml', 'w', encoding='utf-8') as yml_data:
            for i, miu_id in enumerate(ids):
                miu_file_path = file_path + '/' + uri + '.' + miu_id + '.EIS1600'
                yml_header, text = extract_yml_header_and_text(miu_file_path, miu_id, i == 0)
                text_file.write(text)
                yml_data.write(yml_header)



from pathlib import Path

import sys
import os
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from glob import glob
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, read_files_from_readme, write_to_readme, \
    update_texts_fixed_poetry_readme
from eis1600.markdown.methods import convert_to_eis1600_light


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and os.path.isfile(input_arg):
            filepath, fileext = os.path.splitext(input_arg)
            if fileext not in ['.mARkdown', '.inProgess', '.completed']:
                parser.error('You need to input a mARkdown file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to convert mARkdown file(s) to EIS1600TMP file(s).
-----
Give a single mARkdown file as input
or 
Give an input AND an output directory for batch processing.

Use -e <EIS1600_repo> to batch process all mARkdown files in the EIS1600 directory which have not been processed yet.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files which have not been processed yet'
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='MARkdown file to process or input directory with mARkdown files to process if an output directory is '
                 'also given',
            action=CheckFileEndingAction
    )
    arg_parser.add_argument(
            'output', type=str, nargs='?',
            help='Optional, if given batch processes all files from the input directory to the output directory'
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input and not args.output:
        infile = './' + args.input
        if 'data' in infile:
            path = infile.split('data')[0]
        else:
            depth = len(infile.split('/'))
            if depth == 2:
                path = '../../../'
            elif depth == 3:
                path = '../../'
            else:
                path = '../'
        print(f'Convert mARkdown file {infile} to EIS1600TMP file')
        convert_to_eis1600_light(infile, None, verbose)
        infiles = [infile.split('/')[-1]]
        write_to_readme(path, infiles, '# Texts converted into `.EIS1600TMP`\n', '.EIS1600TMP')

    elif args.output:
        input_dir = args.input
        output_dir = args.output
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(f'Convert mARkdown files from {input_dir}, save resulting EIS1600TMP files to {output_dir}')

        infiles = glob(input_dir + '*.mARkdown')
        if not infiles:
            print(
                    'The input directory does not contain any mARkdown files to process'
            )
            sys.exit()

        # Check if output directory exists else create that directory
        Path(output_dir).mkdir(exist_ok=True, parents=True)

        params = [(infile, output_dir, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(convert_to_eis1600_light, params).get()
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(f'Update list of texts with automatically fixed poetry in the README')
        update_texts_fixed_poetry_readme(input_dir, '# Texts with fixed poetry\n')
        print(f'List of texts with automatically fixed poetry was successfully updated in the README')

        print(f'Convert mARkdown files from the EIS1600 repo (only if there is not an EIS1600TMP file yet)')
        files_list = read_files_from_readme(input_dir, '# Texts with fixed poetry\n')
        infiles = get_files_from_eis1600_dir(
                input_dir, files_list, ['mARkdown', 'inProcress', 'completed'], 'EIS1600*'
        )
        if not infiles:
            print(
                    'There are no more mARkdown files to process'
            )
            sys.exit()

        params = [(infile, None, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(convert_to_eis1600_light, params).get()

        write_to_readme(input_dir, infiles, '# Texts converted into `.EIS1600TMP`\n', '.EIS1600TMP')
    else:
        print(
                'Pass in a <uri.mARkdown> file to process a single file or use the -e option for batch processing'
        )
        sys.exit()

    print('Done')
